import os
import json
from prefect import task, flow 
from wopo.prompts import Prompts
from wopo.agent import Agent 
import pandas as pd
from tqdm import tqdm  

class WOPO:
    def __init__(self, prompt, data, text_gen_func = None, process_verbose: bool = False):
        self.agents = [Agent(prompt, datapoint["context"], datapoint["output"], text_gen_func) for datapoint in data]
        self.prompter = Prompts()
        
        assert text_gen_func is not None, "Text Generation Function not provided"
        self.text_gen = text_gen_func

        self.optimal_prompt = None 

    def run_optimisation(self, mode: str = "brute_force", num_iters: int = 1, num_steps_per_iter: int = 1, top_k: int = 1, retries_on_failure: int = 3, retry_delay_seconds: int = 15): 

        results = {}
        agent_states = {}

        @task(retries = retries_on_failure, retry_delay_seconds = retry_delay_seconds) 
        def run_one_step_for_agent(a):
            a.prompt_step()
                    
        @task(retries = retries_on_failure, retry_delay_seconds = retry_delay_seconds)
        def check_correctness_for_agent(a):
            return a.log_last_step()

        @flow(retries = retries_on_failure, retry_delay_seconds = retry_delay_seconds) 
        def run_the_step():
            run_one_step_for_agent.map(self.agents)

        @flow(retries = retries_on_failure, retry_delay_seconds = retry_delay_seconds)
        def correctness():
            temp = check_correctness_for_agent.map(self.agents)
            return temp
   
        def consolidate_scores(agent_scores):
            df = pd.DataFrame.from_dict(agent_scores)
            avg_score = df["score"].mean()
            min_score = df["score"].min()
            max_score = df["score"].max()

            results = {
                "avg": avg_score,
                "max": max_score,
                "min": min_score
            }

            df = df.nlargest(top_k, "score")
            df = df["prompt"]
            prompt_list = df.tolist()
            if len(prompt_list) > 1:
                consolidate_prompt = self.prompter.consolidation_prompt(prompt_list)
                consolidate_prompt = self.text_gen(consolidate_prompt)
            else:
                consolidate_prompt = prompt_list[0]

            return consolidate_prompt, results


        for i in tqdm(range(num_iters)):
            
            for step in range(num_steps_per_iter):
                run_the_step()
        
            correctness_scores = correctness()
            correctness_scores = [a.result() for a in correctness_scores]
            consolidated_prompt, step_results = consolidate_scores(correctness_scores)
            
            agent_states[i] = [agent.logs() for agent in self.agents]

            self.agents = [agent.flush(consolidated_prompt) for agent in self.agents]
            
            results[i] = step_results

        self.optimal_prompt = consolidated_prompt
        return consolidated_prompt, results, agent_states

    def run_single_chain_optimisation(self, num_iters: int = 1, stop_at_score: int = 100):
        assert len(self.agents) == 1, "There is more than one agent. You should use the YAPO.run_optimisation() function instead"
        
        agent = self.agents[0]

        for i in tqdm(range(num_iters)):
            agent.prompt_step()
            if agent.log_last_step()["score"] >= stop_at_score:
                print(f"Early stopping at step {i}")
                break
        self.optimal_prompt = agent.prompt 
        return agent.prompt

    def run_test(self, data, save_location: str = None):
        assert self.optimal_prompt is not None, "Optimal prompt not provided or found"
        df_data = []

        for d in data:
            prompt = self.prompter.execute_prompt(self.optimal_prompt, d["context"])
            extracted_data = self.text_gen(prompt)

            correctness_prompt = self.prompter.correctness_prompt(d["output"], extracted_data)
            correctness_info = self.text_gen(correctness_prompt)
            correctness_info = json.loads(correctness_info)

            df_data_dict = {
                "correct_output": d["output"],
                "received_output": extracted_data,
                "score": correctness_info["correctness"],
                "feedback": correctness_info["explanation"]
            }
            df_data.append(df_data_dict)
        
        df = pd.DataFrame.from_dict(df_data)
        df.to_csv(save_location)

        return df







from wopo.prompts import Prompts
import json 

class Agent:
    def __init__(self, prompt: str = None, data: str = None, correct_output: str = None, text_gen = None):
        self.prompt = prompt
        self.original_prompt = prompt 
        self.data = data 
        self.correct_output = correct_output
        self.prompt_funcs = Prompts()
        self.last_output = None
        self.last_feedback = None
        self.step = 0
        self.state = []
        self.text_gen = text_gen

    
    def update_state(self, prompt: str = None, output: str = None, feedback: str = None, score: int = None):
        self.prompt = prompt
        self.last_output = output
        self.last_feedback = feedback        
        update_dict = {
            self.step: {
                "prompt": prompt,
                "output": output,
                "feedback": feedback,
                "score": score
            }
        }
       
        self.state.append(update_dict)

        self.step += 1
 
    def generate_prompt_and_output(self, prompt):
        receive_prompt = self.text_gen(prompt)

        clean_prompt = self.prompt_funcs.clean_prompt(receive_prompt)
        clean_prompt = self.text_gen(clean_prompt)    
        
        receive_output = self.prompt_funcs.execute_prompt(clean_prompt, self.data)
        receive_output = self.text_gen(receive_output)

        return clean_prompt, receive_output


    def prompt_step(self):
        if self.step == 0:
            design_prompt = self.prompt_funcs.initial_prompt(self.prompt, self.data)
            prompt, output = self.generate_prompt_and_output(design_prompt)
            correctness_check = self.check_correctness(output)
            feedback = correctness_check["explanation"]
            score = correctness_check['correctness']
            self.update_state(prompt, output, feedback, score)
            
        else:
            design_prompt = self.prompt_funcs.update_prompt_with_feedback(self.prompt, self.data, self.correct_output, self.last_output, self.last_feedback)
            prompt, output = self.generate_prompt_and_output(design_prompt)
            correctness_check = self.check_correctness(output)
            feedback = correctness_check["explanation"]
            score = correctness_check['correctness']

            last_step = self.log_last_step()
            if last_step["score"] <= score:
                self.update_state(prompt, output, feedback, score)

    
    def check_correctness(self, output):
        correctness_prompt = self.prompt_funcs.correctness_prompt(self.correct_output, output)
        response = self.text_gen(correctness_prompt)
        dictified = json.loads(response)
        return dictified

    def log_last_step(self):
        return self.state[-1][self.step - 1]
    
    def logs(self):
        return self.state

    def flush(self, new_prompt: str = None):
        if new_prompt is None: 
            self.prompt = self.original_prompt
        else: 
            self.prompt = new_prompt

        self.state = []
        self.last_feedback = None
        self.last_output = None 
        self.step = 0

        return self


class Prompts:

    def initial_prompt(self, prompt: str, data: str = None):
        prompt = f"""
        Design a prompt in the following format: "Acting as a [role], [task]". The task is being performed on the data called context. Be as clear and detailed in the steps as possible. The entire prompt must be a single paragraph
        The task is as follows: {prompt}
        And the context for the task is as follows: {data}
        """
        return prompt

    def update_prompt_with_feedback(self, prompt: str, data: str, expected_output: str, received_output: str, feedback: str):
        prompt = f"""
        You are a prompt engineer. When the following prompt was used on the specific data called data, we receieved output received_output. However, the expected output was expected_output. Rewrite the prompt to better align the data with expected_output based on the feedback provided. The expected prompt should contain instructions in English and should not involve any code

        prompt: {prompt}
        data: {data}
        expected_output: {expected_output}
        received_output: {received_output}
        feedback: {feedback}
        """
        return prompt
    
    def execute_prompt(self, prompt:str, data: str):
        prompt = f"""
        {prompt}

        {data}
        """
        return prompt

    def clean_prompt(self, prompt: str):
        prompt = f"""
        Given this prompt, only retain steps relevant to the prompt

        prompt: {prompt}
        """

        return prompt

    def correctness_prompt(self, expected_output: str, received_output: str):
        prompt = f"""
        Compare these two answers. Give them a score between 0 and 100 for similarity. The response must be a JSON string with keys correctness: score, and explanation: contains explanation for why these two are different

        expected_output: {expected_output}

        received_output: {received_output}
        """
        return prompt

    def consolidation_prompt(self, prompts_list):
        prompts = "\n\n".join(prompts_list)

        prompt = f"""
        You are a prompt engineer. Extract all unique points from all the prompts. Then combine all these prompts into one single prompt that is a single paragraph. Only return the combined prompt

        {prompts}
        """

        return prompt


from setuptools import setup, find_packages
import os

VERSION = '0.0.1'
DESCRIPTION = 'wordlabs open prompt optimiser: Automatic prompt optimisation'
long_description = 'Non parametric plug and play prompt optimisation'

# Setting up
setup(
    name="wopo",
    version=VERSION,
    author="wordlabs.io",
    author_email="<tanishk.kithannae@wordlabs.io>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=['tqdm', 'pandas'],
    keywords=['python', 'nlp', 'prompt', 'optimisation'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)